/*
This is Genetic algorithms!
An ants colony designed to cross a John Muir trail
by process of evolution the process being carried out
in the way of Genetic algorithms.

*/

/*
Genetic Algorithms : applied in this case is as follows:
Steps:
1. Start a Population of 1000 ants with random genes of length 480 bits.
2. Let them try to cross the John Muir trail considering their genes to
	be examples of Finite State Automata. Rules will have to be established
	that generate one out of 4 states from processing the incoming signal
	i.e the state of the block in front of it with the current state of gene.
	The 4 states that can be generated are move forward, face right, face left,
	or stay in place. In the process, the genes are to be altered.
3. A method to serve as a measure of fitness is applied and the fitness values
	of all the ants is determined. The fitness is derived from how far the ant
	travels the Muir trail.
4. Top 10% of the ants are retained and are copied over all the other ants.
	That means you have a thousand ants whose average fitness is more than
	the previuos generation.
5. They are paired randomly and mated. Mating is done by performing a
	exchange of gene after a randomly chosen crossover point. After all the
	ants have been mated, mutation is performed with a rate like 1 in hundred
	or something. Mutation is nothing but toggling of bits.The resulting 1000
	ants are nothing but the next generation of ants.
6. Now start again from step 2 to go on generating new generations while the
	ants get fitter and better.

	Interesting,isnt it?
*/
#include <STDIO.H>
#include <CONIO.H>
#include <STDLIB.H>
#include <TIME.H>
#include <STRING.H>
#include <ASSERT.h>

void error(char* str);
void create_first_gen();
int test_ant_fitness(char*);
void test_ants_fitness(int);
void show_statistics(int);
int fill_trail();
void generate_next_gen(int);
void mate(int);
void crossover(char*, char*);
int next(int, int);
inline int get_bit(char* gene, int t);
inline void put_bit(char*, int, int);
int yesno();
void mutate(int);

const int YES = 1, NO = 0;

int total_ants;
int gene_length;
int no_of_generations;

int iofiles_ver;		//version of format of input output files
int high_num;		//highest number, target block number
char trail_filename[100];
int grid[32][32];
char synch_char = 0;
int file_offset;
int mask[] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };

void main()
	{
	puts("\n\n\n\n");

	puts("Enter the number of ants in a generation(divisible by 10: ");
	scanf("%d", &total_ants);
	total_ants = (total_ants/10)*10;

	puts("\nEnter the gene length in bytes: ");
	scanf("%d", &gene_length);

	puts("\nEnter the number of generations(0 if unsure): ");
	scanf("%d", &no_of_generations);

	puts("\nEnter the version to use for reading and writing files(like 1 or 2):");
	scanf("%d", &iofiles_ver);

	if (iofiles_ver != 1)
		error ("Currently supports version 1 only");

	puts("\nEnter the file containing the trail(JOHNMUIR contains John Muir Trail)");
	scanf("%s", trail_filename);
	if (strchr(trail_filename,'.')==NULL) strcat(trail_filename,".GRD");

	if (fill_trail())
		error("Error reading file! Use the provided GRDMAKE to make the trail.");

	//step 1:
	printf("Generating Generation 0...");
	create_first_gen();
	printf("complete.\n");

	int generation = 0;
	while(1)
		{
		printf("\nTesting ants for fitness in generation no: %d...\n", generation);

		test_ants_fitness(generation);

		printf("\nComplete.");

		printf("\nStatistics for this generation no. %d:", generation);
		show_statistics(generation);
		printf("\nComplete.");

		if (no_of_generations!=0)
			no_of_generations--;
		else
			{
			printf("\n%d generation complete. Do another generation (Y/N)?", generation);
			if (yesno() == NO) break;
			}

		generation++;

		printf("\n\nGenerating generation no:%d...", generation);
		generate_next_gen(generation);
		printf("complete.");

		}

	getch();

	}

void create_first_gen()
	{
	FILE* fp;

	if ( (fp = fopen ( "GEN0.ANT", "wb")) == NULL)
		error("Cannot create file GEN0.ANT");

	/*
	int length = strlen(trail_filename);

	fwrite(&iofiles_ver, sizeof(int), 1, fp);
	fwrite(&total_ants, sizeof(int), 1, fp);
	fwrite(&gene_length, sizeof(int), 1, fp);
	fwrite(&length, sizeof(int), 1, fp);
	fwrite(trail_filename, length, 1, fp);
	file_offset = 2 + 2 + 2 + 2 + length;
	*/
	file_offset = 0;

	randomize();

	int i,j;
	for ( i=0; i<total_ants; i++)
		{
		for ( j=0; j<gene_length; j++)
			fputc(random(256), fp);
		fputc(synch_char, fp);	//synchronization byte
		}

	fclose (fp);
	}

int fill_trail()
	{
	int i,j;
	FILE* fp = fopen ( trail_filename, "rb");

	if (fp==0)	return 1;		//error

	fseek(fp, 0,SEEK_END);
	long int fl = ftell(fp);
	if (fl!=1024) return 2;		//error
	fseek(fp, 0, SEEK_SET);

	for(i=0; i<=31; i++)
		for (j=0; j<=31; j++)
			if ( fread(&grid[i][j], sizeof(char), 1, fp) != 1)
				{
				fclose(fp);
				return 3;       //error
				}
	fclose(fp);

	high_num=0;
	for(i=0; i<=31; i++)
		for ( j=0; j<=31; j++)
			if (grid[i][j] > high_num)
				high_num = grid[i][j];

	return 0;
	}

void test_ants_fitness(int generation)
	{
	char ifilename[13];
	char ofilename[13];

	sprintf(ifilename, "GEN%d.ANT", generation);
	sprintf(ofilename, "GEN%d.FIT", generation);

	FILE *ifp, *ofp;
	ifp = fopen(ifilename, "rb");
	ofp = fopen(ofilename, "wb");
	if (ifp==NULL || ofp==NULL)
		error("File i/o error! Possibly a BUG in Program! 1");

	fseek(ifp, file_offset, SEEK_SET);		//skip info

	char* gene = (char*) malloc(gene_length);
	for (int i=0; i<total_ants; i++)
		{
		cprintf ("\r Processing ant no. %d...", i+1);
		if ( fread(gene, gene_length, 1, ifp) != 1)
			error("File i/o error! Possibly a BUG in Program! 2");

		if (fgetc(ifp) != synch_char)
			error("File Synchronizatin error! Possibly a Bug in Program");

		int fitness = test_ant_fitness(gene);

		fwrite(&fitness, sizeof(int), 1, ofp);
		}

	fclose(ifp);
	fclose(ofp);
	}

void error(char* str)
	{
	puts("\n");
	puts(str);
	getch();
	exit(1);
	}


int test_ant_fitness(char* gene)
	{
	int env_input(int, int,int);
	int process(char*, int);

	int ant_pos = 0;
	int ant_dir = 0;
	int time_step = 1;
	int max_num_reached = 0;
	int score=0;

	while(time_step<=200)
		{
		int decision = process(gene, env_input(ant_pos, ant_dir, max_num_reached) );

		switch(decision)
			{
			case 0:
				//do nothing
				break;
			case 1:
				//turn left
				ant_dir = (ant_dir+1) % 4;
				break;
			case 2:
				//turn right
				ant_dir = (ant_dir+3) % 4;
				break;
			case 3:
				//move forward
				ant_pos = next ( ant_pos, ant_dir );

				if (grid[ant_pos/32][ant_pos%32]==max_num_reached+1)
					{		//update score
					max_num_reached++;
					score += 10;
					if (max_num_reached==89) goto out_of_loop;
					}
				break;

			}

		score--;
		time_step++;
		}

	out_of_loop:

	return score;
	}

int env_input(int pos, int dir, int max_num_reached)
	{
	int t = next (pos, dir);

	return ( (grid[t/32][t%32]==max_num_reached+1)? 1 : 0 );

	}

int process(char* gene, int input)
	{
	/*
	THE CORE ROUTINE :
		It considers the gene as a finite state automata
		it takes in the state of the block in front of the ant
		and processes it with the gene, resulting into a 0 to 3 answer.
		0=do nothing, 1=turn left, 2 = turn right, 3 = move forward a step
		In the process, it also alters the gene slightly so that it is dynamic
		in nature during the course of the trail.
		The procedure followed here is:
		The answer consists of 2 bits.

		How to get the 1st bit:
			Check the 1st bit in the gene. If it is 0 then AND the input bit
			(input bit is whether the block in front is next in path or not) with
			the following bit in the gene. If it is 1 then OR the input bit
			with the following bit in the gene. Now take this result as the
			input bit and repeat the process for the rest of the gene.

		How to get the 2nd bit:
			The same procedure is adopted except that 0 means OR and 1 means
		AND.

		How the gene gets altered in the process:
			The gene gets altered during the finding of the 2nd bit. As we move
		from left to right, one should replace the active phrases 01 and 10
		with passive ones 00 and 11. For this the method devised is to make
		a cycle: 01->00->10->11->01. If the flowing bit is 0, then move the
		phrase up the cycle. If the flowing bit is 1, then do nothing.

	*/

	int to_return=0;
	int dec_bit, op_bit;
	int flow_bit, t;

	//calculation of bit no. 1
	flow_bit = input;
	t = 0;		//counter

	while( t < gene_length*8 )
		{
		dec_bit = get_bit(gene, t);
		t++;
		op_bit = get_bit(gene, t);
		t++;

		if (dec_bit)	//if dec_bit = 1 then OR with op_bit
			flow_bit |= ( (op_bit)? 1 : 0 );
		else				//if dec_bit = 0 then AND with op_bit
			flow_bit &= ( (op_bit)? 1 : 0 );

		}

	assert(flow_bit==0 || flow_bit==1);

	if (flow_bit==1) to_return = 2;//to_Return has this bit as 2nd last bit

	//calculation of 2nd bit
	flow_bit = input;
	t = 0;		//counter reset

	while( t < gene_length*8 )
		{
		int old_flow_bit = flow_bit;  //save current state for alteration of
								//gene later in this loop

		dec_bit = get_bit(gene, t);
		t++;
		op_bit = get_bit(gene, t);
		t++;


		if (dec_bit)	//if dec_bit = 1 then AND with op_bit
			flow_bit &= op_bit;
		else				//if dec_bit = 0 then OR with op_bit
			flow_bit |= op_bit;

		//alter the current gene
		//if flow_bit = 0 then 01->00->10->11->01
		if (old_flow_bit == 0)
			{
			int temp = dec_bit*2 + op_bit;
			int t1,t2;
			switch(temp)  //01 -> 00 -> 10 --> 11 --> 01
				{
				case 1:			t1 = 0; t2 = 0; break;
				case 0:			t1 = 1; t2 = 0; break;
				case 2:			t1 = 1; t2 = 1; break;
				case 3:			t1 = 0; t2 = 1; break;
				}
			put_bit(gene, t-2, t1);
			put_bit(gene, t-1, t2);
			}
		}


	assert(flow_bit==0 || flow_bit==1);

	if (flow_bit) to_return++;

	return to_return;

	}

void put_bit(char* gene, int t, int bit)
	{
	if (bit==1)
		gene[t/8] |= mask[t%8];		//sets the required bit
	else
		gene[t/8] &= (~mask[t%8]); //resets the bit
	}

int get_bit(char* gene, int t)
	{        //returns non-zero if bit no.t in gene is 1 otherwise returns 0.
	return ( (gene[t/8] & mask[t%8])? 1:0 );//isnt it nice -- fast and small!
	}

int next(int block, int dir)
	{
	int to_return;

	switch(dir)
		{
		case 0:
			to_return = (block + 32) % (32*32);
			break;

		case 1:
			if ((block%32)!=0) to_return=block-1; else to_return=block+31;
			break;

		case 2:
			to_return = (block + (32*32) - 32) % (32*32);
			break;

		case 3:
			if ((block%32)!=31) to_return=block+1; else to_return=block-31;
			break;
		}
	return to_return;
	}


int yesno()
	{
	char ch;

	while(1)
		{
		ch = getch();
		if ( (ch=='y') || (ch=='Y') ) { cputs("Y\r\n"); return YES; }
		if ( (ch=='n') || (ch=='N') ) { cputs("N\r\n"); return NO; }
		}

	}

void show_statistics(int generation)
	{
	char filename[13];
	int* fitnesses;
	int i;

	sprintf(filename, "GEN%d.FIT", generation);

	FILE* fp = fopen(filename, "rb");

	if (fp==0) error("returning fp null in show_statistics()");

	//fill fitnesses array
	fitnesses = (int*) malloc (sizeof(int) * total_ants);
	if (fitnesses==0) error ("out of memory in show_statistics");

	for (i=0; i<=total_ants-1; i++)
		if (fread ( &fitnesses[i], sizeof(int), 1, fp) != 1)
			error ("Fitness file small in show_Statistics()");

	int high_score, low_score;
	long int total_score;
	low_score = 801; high_score = -200;
	total_score = 0;

	for(i=0; i<total_ants; i++)
		{
		if (fitnesses[i] > high_score) high_score = fitnesses[i];
		if (fitnesses[i] < low_score) low_score = fitnesses[i];
		total_score += fitnesses[i];
		}

	printf("\nHighest score acheived: %d", high_score);
	printf("\nLowest score acheived: %d", low_score);
	printf("\nAverage score acheived: %d", total_score/total_ants);

	free(fitnesses);
	fclose(fp);
	}

void generate_next_gen(int generation)
	{
	//find the 10% superior ants using the prev generation's fitness file
	char filename[13];
	int* fitnesses;
	int i;

	sprintf(filename, "GEN%d.FIT", generation-1);

	FILE* fp = fopen(filename, "rb");

	if (fp==0) error("returning fp null in show_statistics()");

	//fill fitnesses array
	fitnesses = (int*) malloc (sizeof(int) * total_ants);
	if (fitnesses==0) error ("out of memory in show_statistics");

	for (i=0; i<=total_ants-1; i++)
		if (fread ( &fitnesses[i], sizeof(int), 1, fp) != 1)
			error ("Fitness file small in show_Statistics()");

	fclose(fp);

	//now the fitnesses are in memory : fitnesses array
	//so find the top 10% of those
	int no_of_top10_ants = total_ants/10;
	int* top10_ants;

	top10_ants = (int*) malloc(sizeof(int) * no_of_top10_ants);

	int j=0;

	while (j<no_of_top10_ants)
		{
		int high_score=-200;
		int ant_no;

		for(i=0; i<total_ants; i++)
			if (fitnesses[i] > high_score)
				{
				high_score = fitnesses[i];
				ant_no = i;
				}
		top10_ants[j] = ant_no;
		j++;
		fitnesses[ant_no] = -200;
		}

	free(fitnesses);

	//so now we have number of top10 ants in the top10_ants array
	char ifilename[13], ofilename[13];
	FILE *ifp, *ofp;

	sprintf(ifilename, "GEN%d.ANT", generation-1);
	sprintf(ofilename, "GEN%d.TMP", generation);

	ifp = fopen(ifilename, "rb");
	ofp = fopen(ofilename, "wb");
	if(ifp==0 || ofp==0) error("Cant open file in generate_next_gen()");

	char* gene = (char*) malloc(gene_length+1);

	for (i=0; i<no_of_top10_ants; i++)
		{
		fseek(ifp, top10_ants[i]*(gene_length+1), SEEK_SET);

		fread(gene, gene_length+1, 1, ifp);

		for(int k=0; k<10; k++)			//write that ant 10 times
			fwrite (gene, gene_length+1, 1, ofp);
		}

	free(gene);
	free(top10_ants);

	fclose(ifp);
	fclose(ofp);

	mate(generation);
	mutate(generation);
	}

void mate(int generation)
	{
	int i,k;

	char ifilename[13];
	char ofilename[13];

	sprintf(ifilename, "GEN%d.TMP", generation);
	sprintf(ofilename, "GEN%d.ANT", generation);

	FILE* ifp = fopen (ifilename, "rb");
	FILE* ofp = fopen (ofilename, "wb");

	int* done = (int*) malloc(sizeof(int)*total_ants);
	for(i=0; i<total_ants; i++)
		done[i] = 0;

	char* gene1 = (char*) malloc(gene_length+1);
	char* gene2 = (char*) malloc(gene_length+1);

	for(i=0; i<total_ants; i++)
		if (done[i]==0)
			{
			done[i]=1;

			while(1)
				{
				k = i+1  + random(total_ants-1-i);
				if (done[k]==0) break;
				}
			done[k] = 1;

			fseek(ifp, (gene_length+1)*i, SEEK_SET);
			fread(gene1, gene_length, 1, ifp);

			fseek(ifp, (gene_length+1)*k, SEEK_SET);
			fread(gene2, gene_length, 1, ifp);

			crossover(gene1, gene2);

			fwrite(gene1, gene_length, 1, ofp);
			fputc(synch_char, ofp);
			fwrite(gene2, gene_length, 1, ofp);
			fputc(synch_char, ofp);

			}

	free(gene1);
	free(gene2);
	free(done);

	fclose(ifp);
	fclose(ofp);
//	if (unlink(ifilename)!=0) error("Unlink failed");
//	if (rename(ofilename, ifilename) != 0) error("Rename failed");

	}

void crossover(char* gene1, char* gene2)
	{
	int crossover_pt = random(gene_length*8-1);

	for(int i=crossover_pt+1; i<=gene_length*8-1; i++)
		{
		int bit = get_bit(gene1, i);
		put_bit ( gene1, i, get_bit(gene2, i) );
		put_bit ( gene2, i, bit);
		}
	}

void mutate(int generation)
	{
	char ifilename[13];
	char ofilename[13];

	sprintf(ifilename, "GEN%d.ANT", generation);
	sprintf(ofilename, "GEN%d.TMP", generation);

	FILE* ifp = fopen (ifilename, "rb");
	FILE* ofp = fopen (ofilename, "wb");

	char* gene = (char*) malloc(gene_length);

	for ( int i=0; i<total_ants; i++)
		{
		fread(gene, gene_length, 1, ifp);
		fgetc(ifp);

		//mutate gene
		// we will do 5 mutations
		for (int j=0; j<=4; j++)
			{
			int k = random (gene_length*8);

			if (get_bit (gene, k))
				put_bit(gene, k, 0);
			else
				put_bit(gene, k, 1);
			}

		fwrite (gene, gene_length, 1, ofp);
		fputc(synch_char, ofp);

		}

	free(gene);
	fclose(ifp);
	fclose(ofp);

	if (unlink(ifilename)!=0) error("Unlink failed");
	if (rename(ofilename, ifilename) != 0) error("Rename failed");

	}
